//============================================================================
// Name        : ehSpectrumVisualization.cpp
// Author      : Eui-Hyun Kim (uihyun@gmail.com)
// Version     : 2.12.4
// Copyright   : Okuno Laboratory, Kyoto University
// Description :
//============================================================================

#include <iostream>
#include <string>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <libharkio/libharkio.h>
#include <gtk/gtk.h>

using namespace std;
using namespace FD;

class SpectrumVisualization;

DECLARE_NODE(SpectrumVisualization);
/*Node
 *
 * @name SpectrumVisualization
 * @category HARK:BINAURAL+
 * @description Visualization of audio spectrum.
 *
 * @input_name INPUT_AUDIO_SPECTRUM
 * @input_type Matrix<complex<float> >
 * @input_description Windowed spectrum data. A row index is channel, and a column index is frequency.
 *
 * @output_name OUTPUT_AUDIO_SPECTRUM
 * @output_type Matrix<complex<float> >
 * @output_description Windowed and speech-enhanced spectrum data. A row index is channel, and a column index is frequency.
 *
 * @parameter_name WINDOW_NAME
 * @parameter_type string
 * @parameter_value Visualization of audio spectrum
 * @parameter_description Window name of the time-azimuth map [default: Visualization of audio spectrum].
 *
 * @parameter_name VISUALIZATION_TIME_LENGTH
 * @parameter_type float
 * @parameter_value 10.0
 * @parameter_description Visualization window length to show at the same time [default: 10.0 seconds].
 * 
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The length in sample between a frame and a previous frame [default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 END*/

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

gboolean spectrum_visualization_delete_event(GtkWidget* widget, GdkEventConfigure* event, gpointer data);
gboolean spectrum_visualization_configure_event(GtkWidget* widget, GdkEventConfigure* event, gpointer data);
gboolean spectrum_visualization_expose_event(GtkWidget* widget, GdkEventExpose* event, gpointer data);

class ehSpectrumGtk {

public:
  GtkWidget* visualization_window;
  GtkWidget* drawingArea;
  GdkPixmap* pixMap;
  GdkRectangle plotArea;
  GdkGC* plotGc[5];
  int startCount;
  int countLength;
  float advanceTime;
  int samplingRate;

public:
  ehSpectrumGtk()
  {
    this->initialize();
  }

  ~ehSpectrumGtk()
  {

  }

  void initialize()
  {
    visualization_window = 0;
    drawingArea = 0;
    pixMap = 0;
    startCount = 0;
    countLength = 1000;
    advanceTime = 0.025f;
  }

  void visualization_draw_plot_area()
  {

    if (pixMap == 0)
      return;

    float spaceHeight2 = 20;
    float waveHeight = 256;

    gdk_draw_rectangle(pixMap, drawingArea->style->white_gc, true, drawingArea->allocation.x, drawingArea->allocation.y, drawingArea->allocation.width, drawingArea->allocation.height);

    gdk_draw_rectangle(pixMap, drawingArea->style->black_gc, false, plotArea.x, plotArea.y, plotArea.width, waveHeight);
    gdk_draw_rectangle(pixMap, drawingArea->style->black_gc, false, plotArea.x, plotArea.y+waveHeight+spaceHeight2, plotArea.width, waveHeight);

    GdkFont* font = gdk_font_load("fixed");
    char label[32];

    float maxkHz = samplingRate/2000;
    float stepkHz = maxkHz/4;

    snprintf(label, sizeof(label), "Left signal input");
    gdk_draw_string(pixMap, font, drawingArea->style->black_gc, plotArea.x + spaceHeight2/2, plotArea.y + spaceHeight2, label);
    for (float iSamplingRate = 0, adjX = 0; iSamplingRate <= maxkHz; iSamplingRate += stepkHz, adjX++) {
      snprintf(label, sizeof(label), "%2.1fkHz", iSamplingRate);
      gdk_draw_string(pixMap, font, drawingArea->style->black_gc,
		      plotArea.x - 43,
		      plotArea.y + waveHeight - (waveHeight * adjX * 0.25) + 5,
		      label);
    }

    snprintf(label, sizeof(label), "Right signal input");
    gdk_draw_string(pixMap, font, drawingArea->style->black_gc, plotArea.x + spaceHeight2/2, plotArea.y + waveHeight + spaceHeight2 + spaceHeight2, label);
    for (float iSamplingRate = 0, adjX = 0; iSamplingRate <= maxkHz; iSamplingRate += stepkHz, adjX++) {
      snprintf(label, sizeof(label), "%2.1fkHz", iSamplingRate);
      gdk_draw_string(pixMap, font, drawingArea->style->black_gc,
		      plotArea.x - 43,
		      plotArea.y + waveHeight + spaceHeight2 + waveHeight - (waveHeight * adjX * 0.25) + 5,
		      label);
    }


    for (int cIndex = 0; cIndex <= countLength; cIndex += int(countLength/4)) {
      snprintf(label, sizeof(label), "%0.2fs", (startCount+cIndex)*advanceTime );
      gdk_draw_string(pixMap, font, drawingArea->style->black_gc,
		      plotArea.x + plotArea.width * cIndex / countLength - 5,
		      plotArea.y + waveHeight+spaceHeight2+waveHeight+spaceHeight2,
		      label);
    }
  }

  void visualization_update_window()
  {
    GdkRectangle update_rect;
    update_rect.x = drawingArea->allocation.x;
    update_rect.y = drawingArea->allocation.y;
    update_rect.width = drawingArea->allocation.width;
    update_rect.height = drawingArea->allocation.height;
    gtk_widget_draw(drawingArea, &update_rect);
  }
};


gboolean spectrum_visualization_delete_event(GtkWidget* widget, GdkEventConfigure* event, gpointer data)
{
  ehSpectrumGtk *gtk = (ehSpectrumGtk*)data;

  if (gtk->visualization_window) {
    gtk_widget_hide(gtk->visualization_window);
    gtk->visualization_window = 0;
  }
  return true;
}

gboolean spectrum_visualization_configure_event(GtkWidget* widget, GdkEventConfigure* event, gpointer data)
{
  ehSpectrumGtk *gtk = (ehSpectrumGtk*)data;

  // GdkColor : this is a value between 0 and 65535, with 65535 (USHRT_MAX) indicating full intensity.

  if (gtk->pixMap)
    gdk_pixmap_unref(gtk->pixMap);

  gtk->pixMap = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);

  GdkColor red = {0, 0xffff, 0x0000, 0x0000};
  GdkColor green = {0, 0x0000, 0xffff, 0x0000};
  GdkColor blue = {0, 0x0000, 0x0000, 0xffff};
  GdkColor black = {0, 0x0000, 0x0000, 0x0000};
  GdkColor white = {0, 0xffff, 0xffff, 0xffff};
  GdkColor transparentRed = {0, 0xffff, 0xcccc, 0xcccc};

  gdk_colormap_alloc_color(gdk_colormap_get_system(), &red, false, true);
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &green, false, true);
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &blue, false, true);
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &black, false, true);
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &white, false, true);
  gdk_colormap_alloc_color(gdk_colormap_get_system(), &transparentRed, false, true);

  gtk->plotGc[0] = gdk_gc_new(gtk->pixMap);
  gtk->plotGc[1] = gdk_gc_new(gtk->pixMap);
  gtk->plotGc[2] = gdk_gc_new(gtk->pixMap);
  gtk->plotGc[3] = gdk_gc_new(gtk->pixMap);
  gtk->plotGc[4] = gdk_gc_new(gtk->pixMap);

  gdk_gc_set_foreground(gtk->plotGc[0], &red);
  gdk_gc_set_foreground(gtk->plotGc[1], &green);
  gdk_gc_set_foreground(gtk->plotGc[2], &blue);
  gdk_gc_set_foreground(gtk->plotGc[3], &black);
  gdk_gc_set_foreground(gtk->plotGc[4], &white);

  gdk_gc_set_line_attributes(gtk->plotGc[0], 3, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
  gdk_gc_set_line_attributes(gtk->plotGc[1], 3, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
  gdk_gc_set_line_attributes(gtk->plotGc[2], 1, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
  gdk_gc_set_line_attributes(gtk->plotGc[3], 3, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
  gdk_gc_set_line_attributes(gtk->plotGc[4], 1, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);

  gdk_draw_rectangle(gtk->pixMap, widget->style->white_gc, true, 0, 0, widget->allocation.width, widget->allocation.height);

  gtk->plotArea.x = widget->allocation.x + 50;
  gtk->plotArea.y = widget->allocation.y + 10;
  gtk->plotArea.width = widget->allocation.width - 75;
  gtk->plotArea.height = widget->allocation.height - 40;

  gtk->visualization_draw_plot_area();

  return true;
}

gboolean spectrum_visualization_expose_event(GtkWidget* widget, GdkEventExpose* event, gpointer data)
{
  ehSpectrumGtk *gtk = (ehSpectrumGtk*)data;

  gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)], gtk->pixMap,
		  event->area.x, event->area.y,
		  event->area.x, event->area.y,
		  event->area.width, event->area.height);

  return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////







class SpectrumVisualization : public BufferedNode {

  int spectrumInputID;
  int spectrumOutputID;

  string windowName;
  float visualizationWindowLength;
  int advance;
  int samplingRate;

  ehSpectrumGtk gtk;

public:
  SpectrumVisualization(string nodeName, ParameterSet params) : BufferedNode(nodeName, params)
  {
    spectrumInputID = addInput("INPUT_AUDIO_SPECTRUM");
    spectrumOutputID = addOutput("OUTPUT_AUDIO_SPECTRUM");

    visualizationWindowLength = dereference_cast<float>(parameters.get("VISUALIZATION_TIME_LENGTH"));
    windowName = object_cast<String>(parameters.get("WINDOW_NAME"));
    advance = dereference_cast<int>(parameters.get("ADVANCE"));
    samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));

    gtk.startCount = 0;
    gtk.advanceTime = float(advance)/float(samplingRate);
    gtk.countLength = int(visualizationWindowLength/gtk.advanceTime+0.5);
    gtk.samplingRate = samplingRate;

    inputsCache[spectrumInputID].lookBack = gtk.countLength;
    inOrder = true;
  }

  virtual ~SpectrumVisualization()
  {
  }

  void initialize(void)
  {

    this->BufferedNode::initialize();

    gdk_threads_enter();
    gtk_init(0, 0);

    if (!gtk.visualization_window)
      {
	gtk.visualization_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(gtk.visualization_window), windowName.c_str());
	gtk_widget_set_usize(gtk.visualization_window, 620, 572);

	gtk.drawingArea = gtk_drawing_area_new();
	gtk_widget_show(gtk.drawingArea );
	gtk_container_add(GTK_CONTAINER(gtk.visualization_window), gtk.drawingArea );
	gtk_signal_connect(GTK_OBJECT(gtk.visualization_window), "delete_event", (GtkSignalFunc)(spectrum_visualization_delete_event), &gtk);
	gtk_signal_connect(GTK_OBJECT(gtk.drawingArea ), "configure_event", (GtkSignalFunc)(spectrum_visualization_configure_event), &gtk);
	gtk_signal_connect(GTK_OBJECT(gtk.drawingArea ), "expose_event",	(GtkSignalFunc)(spectrum_visualization_expose_event), &gtk);

	gtk_widget_show(gtk.visualization_window);
      }

    gtk.visualization_draw_plot_area();
    gtk.visualization_update_window();
    gdk_threads_leave();
  }


  void calculate(int output_id, int count, Buffer &out)
  {
    // Set input spectrum data
    RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(spectrumInputID, count);
    Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;

    if (count % 20 == 0) {
      gdk_threads_enter();
      if (gtk.visualization_window != 0) {

	if (count > (gtk.startCount+gtk.countLength))
	  gtk.startCount += gtk.countLength / 2;

	gtk.visualization_draw_plot_area();
	visualization_plot(count);
	gtk.visualization_update_window();
      }
      gtk_main_iteration_do(false);
      gdk_flush();
      gdk_threads_leave();
    }

    // Set output spectrum data as input
    (*outputs[spectrumOutputID].buffer)[count] = spectrumInputPtr;
  }

  void visualization_plot(int count)
  {

    float spaceHeight2 = 20.f;
    float waveHeight = 256;
    const int MAX_COLOR_DEPTH = 65535; // unsigned short

    for (int cIndex = gtk.startCount; cIndex <= count; cIndex++) {

      // Set input spectrum data
      RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(spectrumInputID, cIndex);
      Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;

      const int frequencyMoving = int((spectrumInput.ncols()-1)/waveHeight + 0.5);
      int x = gtk.plotArea.x + (cIndex-gtk.startCount) * gtk.plotArea.width / gtk.countLength;
      int nextX = gtk.plotArea.x + (cIndex-gtk.startCount+1) * gtk.plotArea.width / gtk.countLength;

      for (x; x < nextX; x++)
	for (int mic = 0; mic < 2; mic++) {

	  for (int iFrequency = 0; iFrequency <  spectrumInput.ncols()-1; iFrequency+=frequencyMoving) {

	    int y = 10 + waveHeight + mic*(spaceHeight2 + waveHeight) - iFrequency/frequencyMoving;

	    unsigned long powerSpectrum = 0;
	    for (int fm = 0; fm < frequencyMoving; fm++)
	      powerSpectrum += sqrt(spectrumInput(mic, iFrequency+fm).real() * spectrumInput(mic, iFrequency+fm).real() + spectrumInput(mic, iFrequency+fm).imag() * spectrumInput(mic, iFrequency+fm).imag()) / frequencyMoving;

	    GdkGC *gc;
	    // GdkColor : this is a value between 0 and 65535, with 65535 (USHRT_MAX) indicating full intensity.
	    GdkColor color = {0, 0x0000, 0x0000, 0x0000};

	    if (powerSpectrum <= MAX_COLOR_DEPTH)
	      color.red = powerSpectrum;
	    else if (powerSpectrum <= MAX_COLOR_DEPTH*2){
	      color.red = 0xffff; // MAX_COLOR_DEPTH
	      color.green = powerSpectrum - MAX_COLOR_DEPTH;
	    } else if (powerSpectrum <= MAX_COLOR_DEPTH*3){
	      color.red = 0xffff; // MAX_COLOR_DEPTH
	      color.green = 0xffff;
	      color.blue = powerSpectrum - MAX_COLOR_DEPTH*2; // MAX_COLOR_DEPTH
	    } else {
	      color.red = 0xffff; // MAX_COLOR_DEPTH
	      color.green = 0xffff;
	      color.blue = 0xffff;
	    }

	    gc = gdk_gc_new(gtk.pixMap);
	    gdk_colormap_alloc_color(gdk_colormap_get_system(), &color, false, true);
	    gdk_gc_set_rgb_fg_color(gc, &color);
	    //gdk_gc_set_foreground(gc, &color);
	    //gdk_gc_set_line_attributes(gc, 1, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
	    gdk_draw_point(gtk.pixMap, gc, x, y);
	    gdk_colormap_free_colors(gdk_colormap_get_system(), &color, 1);
	    g_object_unref(gc);

	  }

	}
    }
  }

};


